@@ -139,6 +139,10 @@
 #include <pwd.h>
 #include <grp.h>
 
+#if HAVE_ZLIB_H
+#include <zlib.h>
+#endif
+
 #include "dnscap_common.h"
 #include "dnscap.h"
 #define ISC_CHECK_NONE 1
@@ -296,6 +300,7 @@ struct plugin {
 	output_t		(*output);
 	void			(*getopt)(int *, char **[]);
 	void			(*usage)();
+	void (*extension)(int, void*);
 };
 LIST(struct plugin) plugins;
 
@@ -322,6 +327,7 @@ static void network_pkt(const char *, my_bpftimeval, unsigned,
 static output_t output;
 static int dumper_open(my_bpftimeval);
 static int dumper_close(my_bpftimeval);
+static pcap_dumper_t *dnscap_pcap_dump_open(pcap_t *, const char *);
 static void sigclose(int);
 static void sigbreak(int);
 #if HAVE_PTHREAD
@@ -334,6 +340,7 @@ static logerr_t logerr;
 #if !HAVE___ASSERTION_FAILED
 static void my_assertion_failed(const char *file, int line, assertion_type type, const char *msg, int something) __attribute__((noreturn));
 #endif
+static int is_responder(iaddr ia);
 
 
 /* Private data. */
@@ -379,6 +386,7 @@ static int immediate_mode = FALSE;
 static int background = FALSE;
 static char errbuf[PCAP_ERRBUF_SIZE];
 static int v6bug = FALSE;
+static int wantgzip = FALSE;
 static int wantfrags = FALSE;
 static int wanticmp = FALSE;
 static int wanttcp = FALSE;
@@ -837,6 +845,21 @@ help_2(void) {
 		);
 }
 
+static void
+check_gzip() {
+	char *dot = strrchr(dump_suffix, '.');
+	if (dot) {
+		wantgzip = (strcmp(dot, ".gz") == 0) ? TRUE : FALSE;
+	}
+
+#if ! ( HAVE_GZOPEN && (HAVE_FUNOPEN || HAVE_FOPENCOOKIE ))
+	if (wantgzip) {
+		fprintf(stderr, "error: gzip compression requested but not supported\n");
+		exit(1);
+	}
+#endif
+}
+
 static void
 parse_args(int argc, char *argv[]) {
 	mypcap_ptr mypcap;
@@ -1040,6 +1063,7 @@ parse_args(int argc, char *argv[]) {
 		    if (dump_suffix)
 		        free(dump_suffix);
 			dump_suffix = strdup(optarg);
+			check_gzip();
 			break;
 		case 'k':
 			if (dump_type != to_file)
@@ -1163,6 +1187,11 @@ parse_args(int argc, char *argv[]) {
 				}
 				snprintf(sn, sizeof(sn), "%s_usage", p->name);
 				p->usage = dlsym(p->handle, sn);
+				snprintf(sn, sizeof(sn), "%s_extension", p->name);
+				p->extension = dlsym(p->handle, sn);
+				if (p->extension) {
+					(*p->extension)(DNSCAP_EXT_IS_RESPONDER, (void*)is_responder);
+				}
 				snprintf(sn, sizeof(sn), "%s_getopt", p->name);
 				p->getopt = dlsym(p->handle, sn);
 				if (p->getopt)
@@ -1570,6 +1599,16 @@ ia_equal(iaddr x, iaddr y) {
 	return FALSE;
 }
 
+static int
+is_responder(iaddr ia)
+{
+    if (EMPTY(responders))
+        return 1;
+    if (ep_present(&responders, ia))
+        return 1;
+    return 0;
+}
+
 static int
 ep_present(const endpoint_list *list, iaddr ia) {
 	endpoint_ptr ep;
@@ -2416,20 +2455,24 @@ network_pkt(const char *descr, my_bpftimeval ts, unsigned pf,
 		ns_msg msg;
 		ns_sect s;
 
-		match = FALSE;
-		negmatch = FALSE;
+		match = -1;
+		negmatch = -1;
 		if (ns_initparse(dnspkt, dnslen, &msg) < 0) {
 			discard(tcpstate, "failed parse");
 			return;
 		}
-		for (s = ns_s_qd; s < ns_s_max && !match; s++) {
+		/* Look at each section of the message:
+		     question, answer, authority, additional */
+		for (s = ns_s_qd; s < ns_s_max; s++) {
 			char pres[SNAPLEN*4];
 			const char *look;
 			int count, n;
 			ns_rr rr;
 
+			/* Look at each RR in the section (or each QNAME in
+			   the question section). */
 			count = ns_msg_count(msg, s);
-			for (n = 0; n < count && !negmatch; n++) {
+			for (n = 0; n < count; n++) {
 				myregex_ptr myregex;
 
 				if (ns_parserr(&msg, s, n, &rr) < 0) {
@@ -2448,21 +2491,28 @@ network_pkt(const char *descr, my_bpftimeval ts, unsigned pf,
 					look = pres;
 				}
 				for (myregex = HEAD(myregexes);
-				     myregex != NULL && !negmatch;
+				     myregex != NULL;
 				     myregex = NEXT(myregex, link)) {
-					if (((!match) || myregex->not) &&
-					    regexec(&myregex->reg, look,
-						    0, NULL, 0) == 0)
-					{
-						if (myregex->not) {
-							negmatch = TRUE;
-							match = FALSE;
-						} else
-							match = TRUE;
+				    if (myregex->not) {
+				        if (negmatch < 0)
+				            negmatch = 0;
+				    }
+				    else {
+				        if (match < 0)
+				            match = 0;
+				    }
+
+					if (regexec(&myregex->reg, look, 0, NULL, 0) == 0) {
+						if (myregex->not)
+							negmatch++;
+						else
+							match++;
+
 						if (dumptrace >= 2)
 							fprintf(stderr,
-						   "; \"%s\" ~ /%s/ %d %d\n",
+						   "; \"%s\" %s~ /%s/ %d %d\n",
 								look,
+								myregex->not ? "!" : "",
 								myregex->str,
 								match,
 								negmatch);
@@ -2470,7 +2520,11 @@ network_pkt(const char *descr, my_bpftimeval ts, unsigned pf,
 				}
 			}
 		}
-		if (!match) {
+		/*
+		 * Fail if any negative matching or if no match, match can be -1 which
+		 * indicates that there are only negative matching
+		 */
+		if (negmatch > 0 || match == 0) {
 			discard(tcpstate, "failed regex match");
 			return;
 		}
@@ -2660,7 +2714,7 @@ dumper_open(my_bpftimeval ts) {
 	}
 	if (NULL != t) {
 	    if (options.dump_format == pcap) {
-		    dumper = pcap_dump_open(pcap_dead, t);
+		    dumper = dnscap_pcap_dump_open(pcap_dead, t);
 		    if (dumper == NULL) {
 			    logerr("pcap dump open: %s",
 				    pcap_geterr(pcap_dead));
@@ -2957,3 +3011,63 @@ daemonize(void)
 #endif
   logerr("Backgrounded as pid %u", getpid());
 }
+
+#if HAVE_ZLIB_H
+#if HAVE_FUNOPEN
+static int
+gzip_cookie_write(void *cookie, const char *buf, int size) {
+	return gzwrite((gzFile)cookie, (voidpc)buf, (unsigned) size);
+}
+#elif HAVE_FOPENCOOKIE
+static ssize_t
+gzip_cookie_write(void *cookie, const char *buf, size_t size) {
+	return gzwrite((gzFile)cookie, (voidpc)buf, (unsigned) size);
+}
+#endif
+
+static int
+gzip_cookie_close(void *cookie)
+{
+	return gzclose((gzFile)cookie);
+}
+#endif /* HAVE_ZLIB_H */
+
+static pcap_dumper_t *
+dnscap_pcap_dump_open(pcap_t *pcap, const char *path)
+{
+#if HAVE_ZLIB_H
+#if HAVE_GZOPEN
+	if (wantgzip) {
+		FILE *fp = NULL;
+		gzFile z = gzopen(path, "w");
+		if (z == NULL) {
+			perror("gzopen");
+			return NULL;
+		}
+
+#if HAVE_FUNOPEN
+		fp = funopen(z, NULL, gzip_cookie_write, NULL, gzip_cookie_close);
+		if (fp == NULL) {
+			perror("funopen");
+			return NULL;
+		}
+#elif HAVE_FOPENCOOKIE
+		{
+			static cookie_io_functions_t cookiefuncs = {
+				NULL, gzip_cookie_write, NULL, gzip_cookie_close
+			};
+
+			fp = fopencookie(z, "w", cookiefuncs);
+			if (fp == NULL) {
+				perror("fopencookie");
+				return NULL;
+			}
+		}
+#endif
+		return pcap_dump_fopen(pcap, fp);
+	}
+#endif /* HAVE_GZOPEN */
+#endif /* HAVE_ZLIB_H */
+
+	return pcap_dump_open(pcap, path);
+}
